import { Polygon } from './polygon.js';
import { Vector } from './vector.js';

export class Rectangle extends Polygon {
  constructor(position, dimensions) {
    super();
    if (this.vertices.length) {
      throw new Error('Reinitializing geometry is not allowed');
    }
    this.position = Vector.from(position);
    this.dimensions = Vector.from(dimensions);
    // Next point is obtained by moving the specified length along each dimension
    // one at a time, then reversing these movements in the same order.
    let point = Vector.from(this.position);
    for (let dim = 0; dim < point.dim; dim++) {
      this.addVertex(point);
      const increment = Vector.unitVector(dim, point.dim).scale(this.dimensions[dim]);
      point = point.add(increment);
    }
    for (let dim = 0; dim < point.dim; dim++) {
      this.addVertex(point);
      const increment = Vector.unitVector(dim, point.dim).scale(this.dimensions[dim]);
      point = point.subtract(increment);
    }
  }

  get center() {
    return Vector.from(this.dimensions.map((Q, idx) => this.position[idx] + Q / 2));
  }

  doesOverlap(rect) {
    return this.dimensions.every((_, dim) => {
      const thisMin = this.position[dim];
      const thisMax = this.position[dim] + this.dimensions[dim];
      const thatMin = rect.position[dim];
      const thatMax = rect.position[dim] + rect.dimensions[dim];
      return (thisMin <= thatMin && thisMax >= thatMin)
        || (thisMin >= thatMin && thisMin <= thatMax);
    });
  }

  doesContain(point) {
    return this.dimensions.every((_, dim) => {
      const thisMin = this.position[dim];
      const thisMax = this.position[dim] + this.dimensions[dim];
      return point[dim] >= thisMin && point[dim] <= thisMax;
    });
  }

  get aspectRatio() {
    const [width, height] = this.dimensions;
    return height / width;
  }

  lineIntersect(startPoint, direction) {
    const r = Vector.from(direction).normalize();
    let point = Vector.from(startPoint);
    const maxDistance = this.center.subtract(point).magnitude;
    let increment = maxDistance;
    let everInside = false;
    for (let i = 0; i < 15; i++) {
      if (this.doesContain(point)) {
        everInside = true;
        point = point.subtract(r.scale(increment));
      } else {
        point = point.add(r.scale(increment));
      }
      increment /= 2;
    }
    return everInside ? point : null;
  }
}
